import asyncio
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, Iterator, Optional, TypeVar, cast

from aiohttp import ClientError, ClientSession
from yarl import URL

from aiodynamo.types import Timeout

from ..errors import exception_from_response
from .base import HTTP, Headers, RequestFailed

T = TypeVar("T")


@contextmanager
def wrap_errors() -> Iterator[None]:
    try:
        yield
    except (ClientError, asyncio.TimeoutError):
        raise RequestFailed()


@dataclass(frozen=True)
class AIOHTTP(HTTP):
    session: ClientSession

    async def get(
        self, *, url: URL, headers: Optional[Headers] = None, timeout: Timeout
    ) -> bytes:
        with wrap_errors():
            async with self.session.request(
                method="GET", url=url, headers=headers, timeout=timeout
            ) as response:
                if response.status >= 400:
                    raise RequestFailed()
                return await response.read()

    async def post(
        self, *, url: URL, body: bytes, headers: Optional[Headers] = None
    ) -> Dict[str, Any]:
        with wrap_errors():
            async with self.session.request(
                method="POST",
                url=url,
                headers=headers,
                data=body,
            ) as response:
                if response.status >= 400:
                    raise exception_from_response(
                        response.status, await response.read()
                    )
                return cast(
                    Dict[str, Any],
                    await response.json(
                        content_type="application/x-amz-json-1.0", encoding="utf-8"
                    ),
                )



import json
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, Iterator, Optional, cast

from httpx import AsyncClient, HTTPError
from yarl import URL

from aiodynamo.types import Timeout

from ..errors import exception_from_response
from .base import HTTP, Headers, RequestFailed


@contextmanager
def wrap_errors() -> Iterator[None]:
    try:
        yield
    except HTTPError:
        raise RequestFailed()


@dataclass(frozen=True)
class HTTPX(HTTP):
    client: AsyncClient

    async def get(
        self, *, url: URL, headers: Optional[Headers] = None, timeout: Timeout
    ) -> bytes:
        with wrap_errors():
            # FIXME: the `or {}` is not needed but httpx type hints are wrong
            response = await self.client.get(
                str(url), headers=headers or {}, timeout=timeout
            )
            if response.status_code >= 400:
                raise RequestFailed()
            return await response.aread()

    async def post(
        self, *, url: URL, body: bytes, headers: Optional[Headers] = None
    ) -> Dict[str, Any]:
        with wrap_errors():
            # FIXME: the `or {}` is not needed but httpx type hints are wrong
            response = await self.client.post(
                str(url), content=body, headers=headers or {}
            )
            if response.status_code >= 400:
                raise exception_from_response(
                    response.status_code, await response.aread()
                )
            return cast(Dict[str, Any], json.loads(await response.aread()))



from __future__ import annotations

import abc
import decimal
from dataclasses import dataclass, field
from itertools import chain, count
from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union

from .errors import CannotAddToNestedField
from .types import AttributeType, Numeric, ParametersDict
from .utils import low_level_serialize

_ParametersCache = Dict[Tuple[Any, Any], str]

Addable = Union[Numeric, Set[bytes], Set[str], Set[Numeric]]


class ProjectionExpression(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encode(self, params: Parameters) -> str:
        pass

    @abc.abstractmethod
    def __and__(self, other: F) -> ProjectionExpression:
        pass


@dataclass(frozen=True)
class KeyPath:
    root: str
    parts: Sequence[Union[str, int]] = ()


class F(ProjectionExpression):
    """
    This class represents a Field or Attribute in a DynamoDB Item.

    For paths to items, provide them as separate arguments. For example,
    to reference to the second item in a list stored in the key "foo", use
    `F("foo", 1)`.

    It can be used to create Projection Expressions using the & operator,
    for example `F("foo") & F("bar")` is the Projection expression to
    return the two fields "foo" and "bar".

    It is also used to create Condition Expressions. See the various
    methods below.

    Lastly, it is also used to create Update Expressions, see methods
    at the end of the class.
    """

    path: KeyPath

    def __init__(self, root: str, *path: Union[str, int]):
        self.path = KeyPath(root, path)

    def __hash__(self) -> int:
        return hash(self.path)

    # Projection Expressions

    def __and__(self, other: F) -> ProjectionExpression:
        return FieldList([self, other])

    def encode(self, params: Parameters) -> str:
        return params.encode_path(self.path)

    # Condition Expressions

    def does_not_exist(self) -> Condition:
        """
        Checks that a field does not exist.
        """
        return DoesNotExist(self)

    def exists(self) -> Condition:
        """
        Checks that a field exist.
        """
        return Exists(self)

    def attribute_type(self, attribute_type: AttributeType) -> Condition:
        """
        Checks the attribute type of a field.
        """
        return AttributeTypeCondition(self, attribute_type)

    def begins_with(self, substr: str) -> Condition:
        """
        Checks that a field begins with a substring.
        The substring must not be empty.
        Fields matching the substring provided completely are not returned.
        """
        if not substr:
            raise ValueError("Substring may not be empty")
        return BeginsWith(self, substr)

    def between(self, low: Any, high: Any) -> Condition:
        """
        Checks that a field is between two given values.
        """
        return Between(self, low, high)

    def contains(
        self, value: Union[str, bytes, int, float, decimal.Decimal]
    ) -> Condition:
        """
        Checks if a set or list contains a certain value.
        If a string or bytes object is used as a value, they must not be empty.
        """
        if isinstance(value, (bytes, str)) and not value:
            raise ValueError("Value may not be empty")
        return Contains(self, value)

    def size(self) -> Size:
        """
        Allows checking for the item size. Does not return a condition,
        to return a condition, call the appropriate method on the Size
        class instance returned.
        """
        return Size(self)

    def is_in(self, values: Sequence[Any]) -> Condition:
        """
        Checks if the field is in a sequence of values.
        Between one and one hundred values may be provided.
        """
        return In(self, values)

    def gt(self, other: Any) -> Condition:
        """
        Checks if a field is greater than a value.
        """
        return Comparison(self, ">", other)

    def gte(self, other: Any) -> Condition:
        """
        Checks if a field is greater than a value.
        """
        return Comparison(self, ">=", other)

    def lt(self, other: Any) -> Condition:
        """
        Checks if a field is less than a value.
        """
        return Comparison(self, "<", other)

    def lte(self, other: Any) -> Condition:
        """
        Checks if a field is less or equal than a value.
        """
        return Comparison(self, "<=", other)

    def equals(self, other: Any) -> Condition:
        """
        Checks if a field is equal to value.
        """
        return Comparison(self, "=", other)

    def not_equals(self, other: Any) -> Condition:
        """
        Checks if a field is not equal to a value.
        """
        return Comparison(self, "<>", other)

    # Update Expressions

    def set(self, value: Any) -> UpdateExpression:
        """
        Set a field to a value.
        """
        if isinstance(value, (bytes, str)) and not value:
            return UpdateExpression(remove={self})
        return UpdateExpression(set_updates={self: Value(value)})

    def set_if_not_exists(self, value: Any) -> UpdateExpression:
        """
        Set a field to a value if the field does not exist in the item yet.
        """
        if isinstance(value, (bytes, str)) and not value:
            return UpdateExpression()
        return UpdateExpression(set_updates={self: IfNotExists(value)})

    def change(self, diff: Numeric) -> UpdateExpression:
        """
        Change a numeric field by a given value.
        """
        return UpdateExpression(set_updates={self: Modify(diff)})

    def append(self, value: List[Any]) -> UpdateExpression:
        """
        Add items to a list field. Note that the value passed in should be a
        list, not an individual item.
        """
        return UpdateExpression(set_updates={self: Append(list(value))})

    def remove(self) -> UpdateExpression:
        """
        Remove a field.
        """
        return UpdateExpression(remove={self})

    def add(self, value: Addable) -> UpdateExpression:
        """
        Add a value to a field. Only allowed for top level fields.

        For numeric fields add a numeric value.
        For set fields, this will set the field to the union of the existing
        set and the set provided.
        """
        if self.path.parts:
            raise CannotAddToNestedField()
        return UpdateExpression(add={self: value})

    def delete(self, value: Set[Any]) -> UpdateExpression:
        """
        Deletes all items in the set provided from the set stored in this field.
        """
        return UpdateExpression(delete={self: value})


class KeyCondition(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encode(self, params: Parameters) -> str:
        pass


@dataclass(frozen=True)
class HashKey(KeyCondition):
    """
    Used for Key Conditions. To also constrain the Key Condition by the
    range key, create an instance of RangeKey, call a method on it and
    combine the HashKey with the return value of that method call using
    the & operator.
    """

    name: str
    value: Any

    def encode(self, params: Parameters) -> str:
        return f"{params.encode_path(KeyPath(self.name))} = {params.encode_value(self.value)}"

    def __and__(self, other: Condition) -> KeyCondition:
        return HashAndRangeKeyCondition(self, other)


@dataclass(frozen=True)
class RangeKey:
    """
    Can be used to further constrain a Key Condition. Must be used
    together with a HashKey instance.

    The provided methods behave the same as their counterparts in the
    F class.
    """

    name: str

    def begins_with(self, substr: str) -> Condition:
        if not substr:
            raise ValueError("Substring may not be empty")
        return BeginsWith(F(self.name), substr)

    def between(self, low: Any, high: Any) -> Condition:
        return Between(F(self.name), low, high)

    def contains(
        self, value: Union[str, bytes, int, float, decimal.Decimal]
    ) -> Condition:
        return Contains(F(self.name), value)

    def size(self) -> Size:
        return Size(F(self.name))

    def is_in(self, values: Sequence[Any]) -> Condition:
        return In(F(self.name), values)

    def gt(self, other: Any) -> Condition:
        return Comparison(F(self.name), ">", other)

    def gte(self, other: Any) -> Condition:
        return Comparison(F(self.name), ">=", other)

    def lt(self, other: Any) -> Condition:
        return Comparison(F(self.name), "<", other)

    def lte(self, other: Any) -> Condition:
        return Comparison(F(self.name), "<=", other)

    def equals(self, other: Any) -> Condition:
        return Comparison(F(self.name), "=", other)

    def not_equals(self, other: Any) -> Condition:
        return Comparison(F(self.name), "<>", other)


class Parameters:
    def __init__(self) -> None:
        self.names: Dict[str, str] = {}
        self.values: Dict[str, Dict[str, Any]] = {}
        self.names_gen: Iterator[int] = count()
        self.values_gen: Iterator[int] = count()
        self.names_cache: _ParametersCache = {}
        self.values_cache: _ParametersCache = {}

    def encode_name(self, name: Union[str, int]) -> str:
        return self._encode(
            "#n", name, self.names, self.names_gen, self.names_cache, name
        )

    def encode_value(self, value: Any) -> str:
        tag, value = low_level_serialize(value)
        return self._encode(
            ":v",
            {tag: value},
            self.values,
            self.values_gen,
            self.values_cache,
            (tag, value),
        )

    def encode_path(self, path: KeyPath) -> str:
        return "".join(
            chain(
                [self.encode_name(path.root)],
                (
                    f"[{part}]"
                    if isinstance(part, int)
                    else f".{self.encode_name(part)}"
                    for part in path.parts
                ),
            )
        )

    def to_request_payload(self) -> ParametersDict:
        payload: ParametersDict = {}
        if self.names:
            payload["ExpressionAttributeNames"] = self.names
        if self.values:
            payload["ExpressionAttributeValues"] = {
                key: value for key, value in self.values.items()
            }
        return payload

    def _encode(
        self,
        prefix: str,
        thing: Any,
        data: Dict[str, Any],
        index_gen: Iterator[int],
        cache: _ParametersCache,
        cache_key: Any,
    ) -> str:
        try:
            return cache[cache_key]
        except KeyError:
            can_cache = True
        except TypeError:
            can_cache = False

        encoded = f"{prefix}{next(index_gen)}"
        data[encoded] = thing
        if can_cache:
            cache[cache_key] = encoded
        return encoded


@dataclass(frozen=True)
class HashAndRangeKeyCondition(KeyCondition):
    hash_key: HashKey
    range_key_condition: Condition

    def encode(self, params: Parameters) -> str:
        return f"{self.hash_key.encode(params)} AND {self.range_key_condition.encode(params)}"


class Condition(metaclass=abc.ABCMeta):
    def __and__(self, other: Condition) -> Condition:
        return AndCondition(self, other)

    def __or__(self, other: Condition) -> Condition:
        return OrCondition(self, other)

    def __invert__(self) -> Condition:
        return NotCondition(self)

    @abc.abstractmethod
    def encode(self, params: Parameters) -> str:
        pass


@dataclass(frozen=True)
class NotCondition(Condition):
    base: Condition

    def encode(self, params: Parameters) -> str:
        return f"(NOT {self.base.encode(params)})"


@dataclass(frozen=True)
class AndCondition(Condition):
    lhs: Condition
    rhs: Condition

    def encode(self, params: Parameters) -> str:
        return f"({self.lhs.encode(params)} AND {self.rhs.encode(params)})"


@dataclass(frozen=True)
class OrCondition(Condition):
    lhs: Condition
    rhs: Condition

    def encode(self, params: Parameters) -> str:
        return f"({self.lhs.encode(params)} OR {self.rhs.encode(params)})"


@dataclass(frozen=True)
class DoesNotExist(Condition):
    field: F

    def encode(self, params: Parameters) -> str:
        return f"attribute_not_exists({params.encode_path(self.field.path)})"


@dataclass(frozen=True)
class Exists(Condition):
    field: F

    def encode(self, params: Parameters) -> str:
        return f"attribute_exists({params.encode_path(self.field.path)})"


@dataclass(frozen=True)
class AttributeTypeCondition(Condition):
    field: F
    attribute_type: AttributeType

    def encode(self, params: Parameters) -> str:
        return f"attribute_type({params.encode_path(self.field.path)}, {self.attribute_type.value})"


@dataclass(frozen=True)
class BeginsWith(Condition):
    field: F
    substr: str

    def encode(self, params: Parameters) -> str:
        return f"begins_with({params.encode_path(self.field.path)}, {params.encode_value(self.substr)})"


@dataclass(frozen=True)
class Between(Condition):
    field: F
    low: Any
    high: Any

    def encode(self, params: Parameters) -> str:
        return f"{params.encode_path(self.field.path)} BETWEEN {params.encode_value(self.low)} AND {params.encode_value(self.high)}"


@dataclass(frozen=True)
class Contains(Condition):
    field: F
    value: Union[str, bytes, int, float, decimal.Decimal]

    def encode(self, params: Parameters) -> str:
        return f"contains({params.encode_path(self.field.path)}, {params.encode_value(self.value)})"


@dataclass(frozen=True)
class In(Condition):
    field: F
    values: Sequence[Any]

    def encode(self, params: Parameters) -> str:
        encoded_values = [
            encoded
            for encoded in (params.encode_value(value) for value in self.values)
            if encoded is not None
        ]
        if len(encoded_values) < 1:
            raise ValueError("IN Condition requires at least one value")
        if len(encoded_values) > 100:
            raise ValueError("IN Condition may not contain more than 100 values")
        values = ",".join(encoded_values)
        return f"{params.encode_path(self.field.path)} IN ({values})"


@dataclass(frozen=True)
class Comparison(Condition):
    field: F
    operator: str
    other: Any

    def encode(self, params: Parameters) -> str:
        if isinstance(self.other, F):
            other = params.encode_path(self.other.path)
        else:
            other = params.encode_value(self.other)
        return f"{params.encode_path(self.field.path)} {self.operator} {other}"


@dataclass(frozen=True)
class Size:
    field: F

    def equals(self, other: Any) -> Condition:
        return SizeCondition(self.field, "=", other)

    def not_equals(self, other: Any) -> Condition:
        return SizeCondition(self.field, "<>", other)

    def gt(self, other: Any) -> Condition:
        return SizeCondition(self.field, ">", other)

    def gte(self, other: Any) -> Condition:
        return SizeCondition(self.field, ">=", other)

    def lte(self, other: Any) -> Condition:
        return SizeCondition(self.field, "<=", other)

    def lt(self, other: Any) -> Condition:
        return SizeCondition(self.field, "<", other)


@dataclass(frozen=True)
class SizeCondition(Condition):
    field: F
    operator: str
    value: Any

    def encode(self, params: Parameters) -> str:
        if isinstance(self.value, F):
            value = params.encode_path(self.value.path)
        else:
            value = params.encode_value(self.value)
        return f"size({params.encode_path(self.field.path)}) {self.operator} {value}"


class SetAction(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def encode(self, params: Parameters, field: F) -> str:
        pass


@dataclass(frozen=True)
class Value(SetAction):
    value: Any

    def encode(self, params: Parameters, field: F) -> str:
        return params.encode_value(self.value)


@dataclass(frozen=True)
class IfNotExists(SetAction):
    value: Any

    def encode(self, params: Parameters, field: F) -> str:
        return f"if_not_exists({params.encode_path(field.path)}, {params.encode_value(self.value)})"


@dataclass(frozen=True)
class Modify(SetAction):
    change: Numeric

    def encode(self, params: Parameters, field: F) -> str:
        if self.change < 0:
            operator = "-"
            value = self.change * -1
        else:
            operator = "+"
            value = self.change
        return (
            f"{params.encode_path(field.path)} {operator} {params.encode_value(value)}"
        )


@dataclass(frozen=True)
class Append(SetAction):
    values: List[Any]

    def encode(self, params: Parameters, field: F) -> str:
        return f"list_append({params.encode_path(field.path)}, {params.encode_value(self.values)})"


@dataclass(frozen=True)
class UpdateExpression:
    set_updates: Dict[F, SetAction] = field(default_factory=dict)
    remove: Set[F] = field(default_factory=set)
    add: Dict[F, Addable] = field(default_factory=dict)
    delete: Dict[F, Union[Set[bytes], Set[Numeric], Set[str]]] = field(
        default_factory=dict
    )

    def __and__(self, other: UpdateExpression) -> UpdateExpression:
        return UpdateExpression(
            set_updates={**self.set_updates, **other.set_updates},
            remove=self.remove.union(other.remove),
            add={**self.add, **other.add},
            delete={**self.delete, **other.delete},
        )

    def encode(self, params: Parameters) -> Optional[str]:
        bits = []
        if self.set_updates:
            set_expr = ", ".join(
                f"{params.encode_path(field.path)} = {action.encode(params, field)}"
                for field, action in self.set_updates.items()
            )
            bits.append(f"SET {set_expr}")
        if self.remove:
            remove_expr = ", ".join(
                params.encode_path(field.path) for field in self.remove
            )
            bits.append(f"REMOVE {remove_expr}")
        if self.add:
            add_expr = ", ".join(
                f"{params.encode_path(field.path)} {params.encode_value(value)}"
                for field, value in self.add.items()
            )
            bits.append(f"ADD {add_expr}")
        if self.delete:
            del_expr = ", ".join(
                f"{params.encode_path(field.path)} {params.encode_value(value)}"
                for field, value in self.delete.items()
            )
            bits.append(f"DELETE {del_expr}")
        if bits:
            return " ".join(bits)
        return None


@dataclass(frozen=True)
class FieldList(ProjectionExpression):
    fields: List[F]

    def __and__(self, field: F) -> ProjectionExpression:
        return FieldList(self.fields + [field])

    def encode(self, params: Parameters) -> str:
        return ",".join(params.encode_path(field.path) for field in self.fields)



from __future__ import annotations

import abc
import asyncio
import datetime
import random
import time
from dataclasses import dataclass
from enum import Enum, unique
from itertools import count
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    cast,
)

from .errors import Throttled
from .expressions import Parameters, ProjectionExpression
from .types import (
    EncodedGlobalSecondaryIndex,
    EncodedKeySchema,
    EncodedLocalSecondaryIndex,
    EncodedPayPerRequest,
    EncodedProjection,
    EncodedStreamSpecification,
    EncodedThroughput,
    Item,
    TableName,
    Timespan,
)
from .utils import py2dy


@unique
class TimeToLiveStatus(Enum):
    enabling = "ENABLING"
    disabling = "DISABLING"
    enabled = "ENABLED"
    disabled = "DISABLED"


@dataclass(frozen=True)
class TimeToLiveDescription:
    table: str
    attribute: str
    status: TimeToLiveStatus


@dataclass(frozen=True)
class Throughput:
    read: int
    write: int

    def encode(self) -> EncodedThroughput:
        return {
            "ProvisionedThroughput": {
                "ReadCapacityUnits": self.read,
                "WriteCapacityUnits": self.write,
            }
        }


class PayPerRequest:
    MODE: str = "PAY_PER_REQUEST"

    def encode(self) -> EncodedPayPerRequest:
        return {"BillingMode": self.MODE}


class KeyType(Enum):
    string = "S"
    number = "N"
    binary = "B"


@dataclass(frozen=True)
class KeySpec:
    name: str
    type: KeyType


@dataclass(frozen=True)
class KeySchema:
    hash_key: KeySpec
    range_key: Optional[KeySpec] = None

    def __iter__(self) -> Iterator[KeySpec]:
        yield self.hash_key

        if self.range_key:
            yield self.range_key

    def to_attributes(self) -> Dict[str, str]:
        return {key.name: key.type.value for key in self}

    def encode(self) -> List[EncodedKeySchema]:
        return [
            {"AttributeName": key.name, "KeyType": key_type}
            for key, key_type in zip(self, ["HASH", "RANGE"])
        ]


class ProjectionType(Enum):
    all = "ALL"
    keys_only = "KEYS_ONLY"
    include = "INCLUDE"


@dataclass(frozen=True)
class Projection:
    type: ProjectionType
    attrs: Optional[List[str]] = None

    def encode(self) -> EncodedProjection:
        encoded: EncodedProjection = {"ProjectionType": self.type.value}
        if self.attrs:
            encoded["NonKeyAttributes"] = self.attrs
        return encoded


@dataclass(frozen=True)
class LocalSecondaryIndex:
    name: str
    schema: KeySchema
    projection: Projection

    def encode(self) -> EncodedLocalSecondaryIndex:
        return {
            "IndexName": self.name,
            "KeySchema": self.schema.encode(),
            "Projection": self.projection.encode(),
        }


@dataclass(frozen=True)
class GlobalSecondaryIndex(LocalSecondaryIndex):
    throughput: Throughput

    def encode(self) -> EncodedGlobalSecondaryIndex:
        # Cast due to https://github.com/python/mypy/issues/4122
        return cast(
            EncodedGlobalSecondaryIndex,
            {**super().encode(), "ProvisionedThroughput": self.throughput.encode()},
        )


class StreamViewType(Enum):
    keys_only = "KEYS_ONLY"
    new_image = "NEW_IMAGE"
    old_image = "OLD_IMAGE"
    new_and_old_images = "NEW_AND_OLD_IMAGES"


@dataclass(frozen=True)
class StreamSpecification:
    enabled: bool = False
    view_type: StreamViewType = StreamViewType.new_and_old_images

    def encode(self) -> EncodedStreamSpecification:
        spec: EncodedStreamSpecification = {"StreamEnabled": self.enabled}
        if self.enabled:
            spec["StreamViewType"] = self.view_type.value
        return spec


class ReturnValues(Enum):
    none = "NONE"
    all_old = "ALL_OLD"
    updated_old = "UPDATED_OLD"
    all_new = "ALL_NEW"
    updated_new = "UPDATED_NEW"


class TableStatus(Enum):
    creating = "CREATING"
    updating = "UPDATING"
    deleting = "DELETING"
    active = "ACTIVE"


@dataclass(frozen=True)
class TableDescription:
    attributes: Optional[Dict[str, KeyType]]
    created: Optional[datetime.datetime]
    item_count: Optional[int]
    key_schema: Optional[KeySchema]
    throughput: Optional[Throughput]
    status: TableStatus


class Select(Enum):
    all_attributes = "ALL_ATTRIBUTES"
    all_projected_attributes = "ALL_PROJECTED_ATTRIBUTES"
    count = "COUNT"
    specific_attributes = "SPECIFIC_ATTRIBUTES"


@dataclass(frozen=True)
class WaitConfig:
    max_attempts: int
    retry_delay: Timespan

    @classmethod
    def default(cls) -> WaitConfig:
        return WaitConfig(25, 20)

    async def attempts(self) -> AsyncIterator[None]:
        for _ in range(self.max_attempts):
            yield
            await asyncio.sleep(self.retry_delay)


@dataclass(frozen=True)
class MypyWorkaroundThrottleConfigBase:
    time_limit_secs: Timespan = 60


class ThrottleConfig(MypyWorkaroundThrottleConfigBase, metaclass=abc.ABCMeta):
    @classmethod
    def default(cls) -> ThrottleConfig:
        return ExponentialBackoffThrottling()

    @abc.abstractmethod
    def delays(self) -> Iterable[float]:
        raise NotImplementedError()

    async def attempts(self) -> AsyncIterable[None]:
        deadline = time.monotonic() + self.time_limit_secs
        for delay in self.delays():
            yield
            if time.monotonic() > deadline:
                raise Throttled()
            await asyncio.sleep(delay)


@dataclass(frozen=True)
class DecorrelatedJitterThrottling(ThrottleConfig):
    base_delay_secs: Timespan = 0.05
    max_delay_secs: Timespan = 1

    def delays(self) -> Iterator[Timespan]:
        current_delay_secs = self.base_delay_secs
        while True:
            current_delay_secs = min(
                self.max_delay_secs,
                random.uniform(self.base_delay_secs, current_delay_secs * 3),
            )
            yield current_delay_secs


@dataclass(frozen=True)
class ExponentialBackoffThrottling(ThrottleConfig):
    base_delay_secs: Timespan = 2
    max_delay_secs: Timespan = 20

    def delays(self) -> Iterable[float]:
        for attempt in count():
            yield min(
                random.random() * (self.base_delay_secs ** attempt), self.max_delay_secs
            )


@dataclass(frozen=True)
class Page:
    items: List[Item]
    last_evaluated_key: Optional[Dict[str, Any]]

    @property
    def is_last_page(self) -> bool:
        return self.last_evaluated_key is None


@dataclass(frozen=True)
class BatchGetRequest:
    keys: List[Item]
    projection: Optional[ProjectionExpression] = None

    def to_request_payload(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "Keys": [py2dy(key) for key in self.keys],
        }
        if self.projection:
            params = Parameters()
            payload["ProjectionExpression"] = self.projection.encode(params)
            payload.update(params.to_request_payload())
        return payload


@dataclass(frozen=True)
class BatchGetResponse:
    items: Dict[TableName, List[Item]]
    unprocessed_keys: Dict[TableName, List[Item]]


@dataclass(frozen=True)
class BatchWriteRequest:
    keys_to_delete: Optional[List[Item]] = None
    items_to_put: Optional[List[Item]] = None

    def to_request_payload(self) -> List[Dict[str, Any]]:
        payload: List[Dict[str, Any]] = []
        if self.keys_to_delete:
            payload.extend(
                {"DeleteRequest": {"Key": py2dy(key)}} for key in self.keys_to_delete
            )
        if self.items_to_put:
            payload.extend(
                {"PutRequest": {"Item": py2dy(item)}} for item in self.items_to_put
            )
        return payload


@dataclass(frozen=True)
class BatchWriteResult:
    undeleted_keys: List[Item]
    unput_items: List[Item]



import base64
import datetime
import decimal
import logging
from collections import abc as collections_abc
from functools import reduce
from typing import Any, Callable, Dict, List, Mapping, Set, Tuple, TypeVar, Union

from .types import DynamoItem, Item, NumericTypeConverter

T = TypeVar("T")


logger = logging.getLogger("aiodynamo")


def py2dy(data: Union[Item, None]) -> Union[DynamoItem, None]:
    if data is None:
        return data

    return serialize_dict(data)


def dy2py(data: DynamoItem, numeric_type: NumericTypeConverter) -> Item:
    return {key: deserialize(value, numeric_type) for key, value in data.items()}


def deserialize_simple_types(val: T, _: NumericTypeConverter) -> T:
    return val


def deserialize_binary(val: str, _: NumericTypeConverter) -> bytes:
    return base64.b64decode(val)


def deserialize_string_set(val: List[str], _: NumericTypeConverter) -> Set[str]:
    return set(val)


def deserialize_binary_set(val: List[str], _: NumericTypeConverter) -> Set[bytes]:
    return {base64.b64decode(v) for v in val}


def deserialize_null(_: None, __: NumericTypeConverter) -> None:
    return None


def deserialize_number(val: str, numeric_type: NumericTypeConverter) -> Any:
    return numeric_type(val)


def deserialize_number_set(
    val: List[str], numeric_type: NumericTypeConverter
) -> Set[T]:
    return {numeric_type(v) for v in val}


def deserialize_list(val: List[Any], numeric_type: NumericTypeConverter) -> List[Any]:
    return [deserialize(v, numeric_type) for v in val]


def deserialize_map(
    val: Dict[str, Any], numeric_type: NumericTypeConverter
) -> Dict[str, Any]:
    return {k: deserialize(v, numeric_type) for k, v in val.items()}


TAG_DESERIALIZE_MAPPING: Dict[str, Callable[[Any, NumericTypeConverter], Any]] = {
    "S": deserialize_simple_types,
    "SS": deserialize_string_set,
    "N": deserialize_number,
    "NS": deserialize_number_set,
    "B": deserialize_binary,
    "BS": deserialize_binary_set,
    "BOOL": deserialize_simple_types,
    "NULL": deserialize_null,
    "L": deserialize_list,
    "M": deserialize_map,
}


def deserialize(value: Dict[str, Any], numeric_type: NumericTypeConverter) -> Any:
    if not value:
        raise TypeError(
            "Value must be a nonempty dictionary whose key " "is a valid dynamodb type."
        )
    tag, val = next(iter(value.items()))
    try:
        return TAG_DESERIALIZE_MAPPING[tag](val, numeric_type)
    except KeyError:
        raise TypeError(f"Dynamodb type {tag} is not supported")


NUMERIC_TYPES = int, float, decimal.Decimal


def serialize(value: Any) -> Dict[str, Any]:
    """
    Serialize a Python value to a Dynamo Value, removing empty strings.
    """
    tag, value = low_level_serialize(value)
    return {tag: value}


def low_level_serialize(value: Any) -> Tuple[str, Any]:
    if value is None:
        return "NULL", True
    elif isinstance(value, bool):
        return "BOOL", value
    elif isinstance(value, NUMERIC_TYPES):
        return "N", str(value)
    elif isinstance(value, str):
        return "S", value
    elif isinstance(value, bytes):
        return "B", base64.b64encode(value).decode("ascii")
    elif isinstance(value, collections_abc.Set):
        numeric_items, str_items, byte_items, total = reduce(
            lambda acc, item: (
                acc[0] + isinstance(item, NUMERIC_TYPES),
                acc[1] + isinstance(item, str),
                acc[2] + isinstance(item, bytes),
                acc[3] + 1,
            ),
            value,
            (0, 0, 0, 0),
        )
        if numeric_items == total:
            return "NS", [str(item) for item in value]
        elif str_items == total:
            return "SS", [item for item in value]
        elif byte_items == total:
            return (
                "BS",
                [base64.b64encode(item).decode("ascii") for item in value],
            )
        else:
            raise TypeError(
                f"Sets which are not entirely numeric, strings or bytes are not supported. value: {value!r}"
            )
    elif isinstance(value, collections_abc.Mapping):
        return "M", serialize_dict(value)
    elif isinstance(value, collections_abc.Sequence):
        return "L", [item for item in map(serialize, value)]
    else:
        raise TypeError(f"Unsupported type {type(value)}: {value!r}")


def serialize_dict(value: Mapping[str, Any]) -> Dict[str, Dict[str, Any]]:
    return {key: serialize(value) for key, value in value.items()}


def parse_amazon_timestamp(timestamp: str) -> datetime.datetime:
    return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ").replace(
        tzinfo=datetime.timezone.utc
    )



import abc
from typing import Any, Dict, Optional

from yarl import URL

from aiodynamo.types import Timeout

Headers = Dict[str, str]


class RequestFailed(Exception):
    pass


class HTTP(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def get(
        self, *, url: URL, headers: Optional[Headers] = None, timeout: Timeout
    ) -> bytes:
        """
        Make a GET request and return the response as bytes.

        Raise a RequestFailed exception if the request was not successful.
        """
        pass

    @abc.abstractmethod
    async def post(
        self, *, url: URL, body: bytes, headers: Optional[Headers] = None
    ) -> Dict[str, Any]:
        """
        Make a POST request and return the parsed JSON object.

        Raise a RequestFailed exception if the request was not successful.
        """
        pass



from __future__ import annotations

import abc
import asyncio
import configparser
import datetime
import json
import os
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import Callable, Optional, Sequence, TypeVar

from yarl import URL

from .http.base import HTTP, Headers
from .types import Timeout
from .utils import logger, parse_amazon_timestamp


@dataclass(frozen=True)
class Key:
    id: str
    secret: str = field(repr=False)
    token: Optional[str] = field(repr=False, default=None)


class Credentials(metaclass=abc.ABCMeta):
    @classmethod
    def auto(cls) -> ChainCredentials:
        """
        Return the default credentials loader chain.
        """
        return ChainCredentials(
            candidates=[
                EnvironmentCredentials(),
                FileCredentials(),
                ContainerMetadataCredentials(),
                InstanceMetadataCredentials(),
            ]
        )

    @abc.abstractmethod
    async def get_key(self, http: HTTP) -> Optional[Key]:
        """
        Return a Key if one could be found.
        """

    @abc.abstractmethod
    def invalidate(self) -> bool:
        """
        Invalidate the credentials if possible and return whether credentials
        got invalidated or not.
        """

    @abc.abstractmethod
    def is_disabled(self) -> bool:
        """
        Indicate if this credentials provider is disabled. Used by ChainCredentials
        to ignore providers that won't ever find a key.

        If the status could change over the lifetime of a program, return True.
        """


class Disabled(Exception):
    pass


@dataclass(frozen=True)
class StaticCredentials(Credentials):
    """
    Static credentials provided in Python.
    """

    key: Optional[Key]

    async def get_key(self, http: HTTP) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return False


class ChainCredentials(Credentials):
    """
    Chains multiple credentials providers together, trying them
    in order. Returns the first key found. Exceptions are suppressed.
    """

    candidates: Sequence[Credentials]

    def __init__(self, candidates: Sequence[Credentials]) -> None:
        self.candidates = [
            candidate for candidate in candidates if not candidate.is_disabled()
        ]

    async def get_key(self, http: HTTP) -> Optional[Key]:
        for candidate in self.candidates:
            try:
                key = await candidate.get_key(http)
            except:
                logger.exception("Candidate %r failed", candidate)
                continue
            if key is None:
                logger.debug("Candidate %r didn't find a key", candidate)
            else:
                logger.debug("Candidate %r found a key %r", candidate, key)
                return key
        return None

    def invalidate(self) -> bool:
        return any(candidate.invalidate() for candidate in self.candidates)

    def is_disabled(self) -> bool:
        return not self.candidates


class EnvironmentCredentials(Credentials):
    """
    Loads the credentials from the environment.
    """

    key: Optional[Key]

    def __init__(self) -> None:
        try:
            self.key = Key(
                os.environ["AWS_ACCESS_KEY_ID"],
                os.environ["AWS_SECRET_ACCESS_KEY"],
                os.environ.get("AWS_SESSION_TOKEN", None),
            )
        except KeyError:
            self.key = None

    async def get_key(self, http: HTTP) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return self.key is None


# https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html
class FileCredentials(Credentials):
    """
    Loads the credentials from an AWS credentials file
    """

    key: Optional[Key]

    def __init__(
        self,
        *,
        path: Optional[Path] = None,
        profile_name: str = "default",
    ) -> None:
        if path is None:
            path = Path.home().joinpath(".aws", "credentials")
        self.key = None
        if not path.is_file():
            return

        parser = configparser.RawConfigParser()
        try:
            parser.read(path)
        except configparser.Error:
            logger.exception("Found credentials file %r but parsing failed", path)
            return

        try:
            profile = parser[profile_name]
        except KeyError:
            logger.exception(
                "Profile %r not found in credentials file %r", profile_name, path
            )
            return

        try:
            self.key = Key(
                id=profile["aws_access_key_id"],
                secret=profile["aws_secret_access_key"],
            )
        except KeyError:
            logger.exception(
                "Profile %r in %r does not contain credentials", profile_name, path
            )

    async def get_key(self, http: HTTP) -> Optional[Key]:
        return self.key

    def invalidate(self) -> bool:
        return False

    def is_disabled(self) -> bool:
        return self.key is None


class Refresh(Enum):
    required = auto()
    soon = auto()
    not_required = auto()


EXPIRES_SOON_THRESHOLD = datetime.timedelta(minutes=15)
EXPIRED_THRESHOLD = datetime.timedelta(minutes=10)


@dataclass(frozen=True)
class Metadata:
    key: Key
    expires: datetime.datetime

    def check_refresh(self) -> Refresh:
        now = datetime.datetime.now(datetime.timezone.utc)
        if now >= self.expires:
            return Refresh.required
        diff = self.expires - now
        if diff < EXPIRED_THRESHOLD:
            return Refresh.required
        if diff < EXPIRES_SOON_THRESHOLD:
            return Refresh.soon
        return Refresh.not_required


class MetadataCredentials(Credentials, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def fetch_metadata(self, http: HTTP) -> Metadata:
        pass

    async def fetch_with_retry(
        self,
        *,
        http: HTTP,
        max_attempts: int,
        url: URL,
        timeout: Timeout,
        headers: Optional[Headers] = None,
    ) -> bytes:
        for attempt in range(max_attempts):
            try:
                response = await http.get(url=url, headers=headers, timeout=timeout)
            except Exception:
                logger.exception("GET failed")
                continue
            if response:
                logger.debug("fetched metadata %s bytes", len(response))
                return response
        raise TooManyRetries()

    def __post_init__(self) -> None:
        self._metadata: Optional[Metadata] = None
        self._refresher: Optional[asyncio.Task[None]] = None

    async def get_key(self, http: HTTP) -> Optional[Key]:
        if self.is_disabled():
            logger.debug("%r is disabled", self)
            return None
        refresh = self._check_refresh()
        logger.debug("refresh status %r", refresh)
        if refresh is Refresh.required:
            if self._refresher is None:
                logger.debug("starting mandatory refresh")
                self._refresher = asyncio.create_task(self._refresh(http))
            else:
                logger.debug("re-using refresh")
            try:
                await self._refresher
            finally:
                self._refresher = None
        elif refresh is Refresh.soon:
            if self._refresher is None:
                logger.debug("starting early refresh")
                self._refresher = asyncio.create_task(self._refresh(http))
            else:
                logger.debug("already refreshing")
        if self._metadata:
            return self._metadata.key
        return None

    def invalidate(self) -> bool:
        logger.debug("%r invalidated", self)
        self._metadata = None
        return True

    def _check_refresh(self) -> Refresh:
        if self._metadata is None:
            return Refresh.required
        return self._metadata.check_refresh()

    async def _refresh(self, http: HTTP) -> None:
        self._metadata = await self.fetch_metadata(http)
        logger.debug("fetched metadata %r", self._metadata)


T = TypeVar("T")
U = TypeVar("U")


def and_then(thing: Optional[T], mapper: Callable[[T], U]) -> Optional[U]:
    if thing is not None:
        return mapper(thing)
    return thing


# https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-metadata-endpoint.html
@dataclass
class ContainerMetadataCredentials(MetadataCredentials):
    """
    Loads credentials from the ECS container metadata endpoint.
    """

    timeout: Timeout = 2
    max_attempts: int = 3
    base_url: URL = URL("http://169.254.170.2")
    relative_uri: Optional[str] = field(
        default_factory=lambda: os.environ.get(
            "AWS_CONTAINER_CREDENTIALS_RELATIVE_URI", None
        )
    )
    full_uri: Optional[URL] = field(
        default_factory=lambda: and_then(
            os.environ.get("AWS_CONTAINER_CREDENTIALS_FULL_URI", None), URL
        )
    )
    auth_token: Optional[str] = field(
        default_factory=lambda: os.environ.get(
            "AWS_CONTAINER_AUTHORIZATION_TOKEN", None
        )
    )

    @property
    def url(self) -> Optional[URL]:
        if self.relative_uri is not None:
            return self.base_url.with_path(self.relative_uri)
        elif self.full_uri is not None:
            return self.full_uri
        else:
            return None

    def is_disabled(self) -> bool:
        return self.url is None

    async def fetch_metadata(self, http: HTTP) -> Metadata:
        headers = and_then(self.auth_token, lambda token: {"Authorization": token})
        if self.url is None:
            raise Disabled()
        response = await self.fetch_with_retry(
            http=http,
            max_attempts=self.max_attempts,
            url=self.url,
            headers=headers,
            timeout=self.timeout,
        )
        data = json.loads(response)
        return Metadata(
            key=Key(
                id=data["AccessKeyId"],
                secret=data["SecretAccessKey"],
                token=data.get("Token", None),
            ),
            expires=parse_amazon_timestamp(data["Expiration"]),
        )


# https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/instancedata-data-retrieval.html
@dataclass
class InstanceMetadataCredentials(MetadataCredentials):
    """
    Loads credentials from the EC2 instance metadata endpoint.
    """

    timeout: Timeout = 1
    max_attempts: int = 1
    base_url: URL = URL("http://169.254.169.254")
    disabled: bool = field(
        default_factory=lambda: os.environ.get(
            "AWS_EC2_METADATA_DISABLED", "false"
        ).lower()
        == "true"
    )

    def is_disabled(self) -> bool:
        return self.disabled

    async def fetch_metadata(self, http: HTTP) -> Metadata:
        role_url = self.base_url.with_path(
            "/latest/meta-data/iam/security-credentials/"
        )
        role = (
            await self.fetch_with_retry(
                http=http,
                max_attempts=self.max_attempts,
                url=role_url,
                timeout=self.timeout,
            )
        ).decode("utf-8")
        credentials_url = self.base_url.with_path(
            f"/latest/meta-data/iam/security-credentials/{role}"
        )
        raw_credentials = await self.fetch_with_retry(
            http=http,
            max_attempts=self.max_attempts,
            url=credentials_url,
            timeout=self.timeout,
        )
        credentials = json.loads(raw_credentials)
        return Metadata(
            key=Key(
                credentials["AccessKeyId"],
                credentials["SecretAccessKey"],
                credentials.get("Token", None),
            ),
            expires=parse_amazon_timestamp(credentials["Expiration"]),
        )


class TooManyRetries(Exception):
    pass



from __future__ import annotations

import datetime
import hashlib
import hmac
import json
from dataclasses import dataclass, field
from typing import Any, Dict, Mapping, Optional

from yarl import URL

from .credentials import Key

SERVICE = "dynamodb"
CONTENT_TYPE = "application/x-amz-json-1.0"
API_VERSION = "DynamoDB_20120810"
METHOD = "POST"
ALGORITHM = "AWS4-HMAC-SHA256"


def make_default_endpoint(region: str) -> URL:
    return URL.build(scheme="https", host=f"{SERVICE}.{region}.amazonaws.com", path="/")


@dataclass(frozen=True)
class Instant:
    _dt: datetime.datetime

    @classmethod
    def now(cls) -> Instant:
        return cls(datetime.datetime.now(datetime.timezone.utc))

    @property
    def timestamp(self) -> str:
        return self._dt.strftime("%Y%m%dT%H%M%SZ")

    @property
    def date(self) -> str:
        return self._dt.strftime("%Y%m%d")


@dataclass(frozen=True)
class Request:
    url: URL
    headers: Dict[str, str] = field(repr=False)
    body: bytes


def derive_signing_key(key: Key, instant: Instant, region: str) -> bytes:
    tmp_key1 = sha256_hmac(("AWS4" + key.secret).encode("utf-8"), instant.date)
    tmp_key2 = sha256_hmac(tmp_key1, region)
    tmp_key3 = sha256_hmac(tmp_key2, SERVICE)
    return sha256_hmac(tmp_key3, "aws4_request")


def signed_dynamo_request(
    *,
    key: Key,
    payload: Mapping[str, Any],
    action: str,
    region: str,
    endpoint: Optional[URL] = None,
) -> Request:
    instant = Instant.now()
    endpoint = endpoint or make_default_endpoint(region)
    host = endpoint.host
    amz_target = f"{API_VERSION}.{action}"

    canonical_uri = "/"
    canonical_querystring = ""
    canonical_headers = (
        f"content-type:{CONTENT_TYPE}\n"
        f"host:{host}\n"
        f"x-amz-date:{instant.timestamp}\n"
        f"x-amz-target:{amz_target}\n"
    )

    payload_bytes = json.dumps(payload, separators=(",", ":")).encode("utf-8")

    signed_headers = "content-type;host;x-amz-date;x-amz-target"
    payload_hash = hashlib.sha256(payload_bytes).hexdigest()
    canonical_request = (
        f"{METHOD}\n"
        f"{canonical_uri}\n"
        f"{canonical_querystring}\n"
        f"{canonical_headers}\n"
        f"{signed_headers}\n"
        f"{payload_hash}"
    )

    credential_scope = f"{instant.date}/{region}/{SERVICE}/aws4_request"
    request_digest = hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()
    string_to_sign = (
        f"{ALGORITHM}\n"
        f"{instant.timestamp}\n"
        f"{credential_scope}\n"
        f"{request_digest}"
    )

    signing_key = derive_signing_key(key, instant, region)
    signature = hmac.new(
        signing_key, string_to_sign.encode("utf-8"), hashlib.sha256
    ).hexdigest()
    authorization_header = f"{ALGORITHM} Credential={key.id}/{credential_scope}, SignedHeaders={signed_headers}, Signature={signature}"

    headers = {
        "Content-Type": CONTENT_TYPE,
        "X-Amz-Date": instant.timestamp,
        "X-Amz-Target": amz_target,
        "Authorization": authorization_header,
    }
    if key.token:
        headers["X-Amz-Security-Token"] = key.token
    return Request(url=endpoint, headers=headers, body=payload_bytes)


def sha256_hmac(key: bytes, msg: str) -> bytes:
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aiodynamo', 'aiodynamo.http']

package_data = \
{'': ['*']}

install_requires = \
['yarl>=1.4.2,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['typing_extensions>=3.7,<4.0'],
 'aiohttp': ['aiohttp>=3.6.2,<4.0.0'],
 'httpx': ['httpx>=0.15.0,<1.0.0']}

setup_kwargs = {
    'name': 'aiodynamo',
    'version': '21.10',
    'description': 'Asyncio DynamoDB client',
    'long_description': '# AsyncIO DynamoDB\n\n[![CircleCI](https://circleci.com/gh/HENNGE/aiodynamo.svg?style=svg)](https://circleci.com/gh/HENNGE/aiodynamo)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/aiodynamo/badge/?version=latest)](https://aiodynamo.readthedocs.io/en/latest/?badge=latest)\n\nAsynchronous pythonic DynamoDB client; **2x** faster than `aiobotocore/boto3/botocore`.\n\n## Quick start\n\n### With httpx\nInstall this library\n\n`pip install "aiodynamo[httpx]"` or, for poetry users `poetry add aiodynamo -E httpx`\n\nConnect to DynamoDB\n\n```py\nfrom aiodynamo.client import Client\nfrom aiodynamo.credentials import Credentials\nfrom aiodynamo.http.httpx import HTTPX\nfrom httpx import AsyncClient\n\n    async with AsyncClient() as h:\n        client = Client(HTTPX(h), Credentials.auto(), "us-east-1")\n```\n\n### With aiohttp\nInstall this library\n\n`pip install "aiodynamo[aiohttp]"` or, for poetry users `poetry add aiodynamo -E aiohttp`\n\nConnect to DynamoDB\n\n```py\nfrom aiodynamo.client import Client\nfrom aiodynamo.credentials import Credentials\nfrom aiodynamo.http.aiohttp import AIOHTTP\nfrom aiohttp import ClientSession\n\n    async with ClientSession() as session:\n        client = Client(AIOHTTP(session), Credentials.auto(), "us-east-1")\n```\n\n### API use\n\n```py\n        table = client.table("my-table")\n\n        # Create table if it doesn\'t exist\n        if not await table.exists():\n            await table.create(\n                Throughput(read=10, write=10),\n                KeySchema(hash_key=KeySpec("key", KeyType.string)),\n            )\n\n        # Create or override an item\n        await table.put_item({"key": "my-item", "value": 1})\n        # Get an item\n        item = await table.get_item({"key": "my-item"})\n        print(item)\n        # Update an item, if it exists.\n        await table.update_item(\n            {"key": "my-item"}, F("value").add(1), condition=F("key").exists()\n        )\n```\n\n## Why aiodynamo\n\n* boto3 and botocore are synchronous. aiodynamo is built for **asynchronous** apps.\n* aiodynamo is **fast**. Two times faster than aiobotocore, botocore or boto3 for operations such as query or scan.\n* aiobotocore is very low level. aiodynamo provides a **pythonic API**, using modern Python features. For example, paginated APIs are automatically depaginated using asynchronous iterators.\n* **Legible source code**. botocore and derived libraries generate their interface at runtime, so it cannot be inspected and isn\'t typed. aiodynamo is hand written code you can read, inspect and understand.\n* **Pluggable HTTP client**. If you\'re already using an asynchronous HTTP client in your project, you can use it with aiodynamo and don\'t need to add extra dependencies or run into dependency resolution issues.\n\n[Complete documentation is here](https://aiodynamo.readthedocs.io/)\n',
    'author': 'Jonas Obrist',
    'author_email': 'jonas.obrist@hennge.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/HENNGE/aiodynamo',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
